/*
    Author Daniel Rebsdorf
    Website http://drebsdorf.net 
 
    Made for EventGhost/0.5.0-rc6 Webserver Plugin
    This sketch uses cheap ebay $1 RF 433mhz modules to toggle RF Lights

    Connect RF Transmitter to D3 + 3v + Gnd, 
    Connect RF Receiver data to D4 + 3v + Gnd

    Add the WebServer plugin to your EventGhost configuration, and test the connection to the arduino with "Webserver: Send event to another EventGhost" action
    Remember to set port, username and password, in this sketch example its : port 80, username test, password 1234.

    The following format is used for reading values from event ghost
    Event:      decimalCode:13964556,bitLength:24

    This sketch also supports sending events to the EventGhost Webserver plugin
    It will send any incoming RF signals back to Eventghost, so you can also use this sketch to record signals

    *Updates
    *Fixed some json
    *Added pulseLength as an input variable to the arduino
    *
*/

#include <RCSwitch.h>  // Download https://github.com/sui77/rc-switch/releases/latest
#include <ESP8266WiFi.h>  // Boards Url http://arduino.esp8266.com/stable/package_esp8266com_index.json
#include <ESP8266HTTPClient.h>
#include <rBase64.h> 


#define STASSID "WifiNetwork"           //SSID
#define STAPSK  "WifiPassword"          //WPA-PSK
#define HOSTNAME  "ESP8266_433mhz"      //ESP8266 Network Hostname
#define INCOMING_AUTH_REQUIRED true     //Authentication required for arduino
#define INCOMING_AUTH_USERNAME "test"   //Username for arduino
#define INCOMING_AUTH_PASSWORD "1234"   //Password for arduino
#define INCOMING_PORT 80                //Port Arduino listens on
#define OUTGOING_AUTH_REQUIRED true     //If authentication is required by EventGhost Webserver Plugin
#define OUTGOING_AUTH_USERNAME "admin"  //Username of EventGhost Webserver Plugin
#define OUTGOING_AUTH_PASSWORD "admin"  //Password of EventGhost Webserver Plugin
#define OUTGOING_IP "192.168.0.60"      //IP address of EventGhost Webserver Plugin
#define OUTGOING_PORT 8888              //Port of EventGhost Webserver Plugin

const char* ssid = STASSID;
const char* password = STAPSK;
String strDeviceName = HOSTNAME;

//Used for outgoing connections to EventGhost Webserver Plugin
String outgoing_AUTH_USERNAME = OUTGOING_AUTH_USERNAME;
String outgoing_AUTH_PASSWORD = OUTGOING_AUTH_PASSWORD;
String outgoing_IP = OUTGOING_IP;
int outgoing_PORT = OUTGOING_PORT;

// Create an instance of the server
// specify the port to listen on as an argument
WiFiServer server(INCOMING_PORT);

//433Mhz Remote Control Switch
RCSwitch mySwitch = RCSwitch();

void setup() {
  Serial.begin(115200);

  // Connect to WiFi network
  Serial.println();
  Serial.println();
  Serial.print(F("Connecting to "));
  Serial.println(ssid);

  //Set hostname
  WiFi.hostname(strDeviceName);
  WiFi.mode(WIFI_STA);
  WiFi.begin(ssid, password);

  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(F("."));
  }
  Serial.println();
  Serial.println(F("WiFi connected"));

  // Start the server
  server.begin();
  Serial.println(F("Server started"));

  // Print the IP address
  Serial.println(WiFi.localIP());

  //Setup RF Switch pins
  mySwitch.enableReceive(2);  // GPIO-2 = D4, RF Recieve, $1 ebay 433mhz modules
  mySwitch.enableTransmit(0);  // GPIO-0 = D3, RF Transmit, $1 ebay 433mhz modules
  // Optional set pulse length.
  mySwitch.setPulseLength(541);

  // Optional set protocol (default is 1, will work for most outlets)
  // mySwitch.setProtocol(2);

  // Optional set number of transmission repetitions.
  mySwitch.setRepeatTransmit(5);
}



void loop() {

  //Test Example 
  //post("TriggerEvent&prefix=SomeTest&suffix=Event&payload={\'jsondata\': \'json value\'}");
  //delay(2000);
  
  // Check if a client has connected
  WiFiClient client = server.available();
  if (client) {
    Serial.println(F("new client"));
  
    // Read the first line of the request
    String req = client.readStringUntil('\r');
    Serial.println(F("request: "));
    Serial.println(req);
    Serial.println("");
  
    String readString;
    // Read rest of request
    while (client.available()) {
      char c = client.read();
      //store characters to string
      readString += c;
    
    }
    Serial.println(readString);
    Serial.println("");
    
    int authStart = readString.indexOf("Authorization: Basic ") + 21;
    int authEnd = readString.indexOf("Content-Length:") -1;
    String authBase64 = readString.substring(authStart,authEnd);
    Serial.println(authBase64);
    Serial.println("");
  
    bool authenticated = false;
    if(INCOMING_AUTH_REQUIRED) //Auth Required
    {
      rbase64.decode(authBase64);
      String authResult = rbase64.result();
      Serial.println(authResult);
  
      int userStart = 0;
      int userEnd = authResult.indexOf(":");
      String userName = authResult.substring(userStart,userEnd);
      String userPass = authResult.substring(userEnd + 1);
      Serial.println(userName);
      Serial.println(userPass);
  
      if(!userName.equals( INCOMING_AUTH_USERNAME) || !userPass.equals( INCOMING_AUTH_PASSWORD))
      {
        client.print(F("HTTP/1.1 200 OK\r\nContent-Type: application/json; charset=utf-8\r\n\r\n{\"failed\": \"login\"}")); 
        client.stop();
        Serial.println("client disconnected");
      }
      else
      {
        // Authenticated
        authenticated = true;
        Serial.println("client authorized");
      }
      
    }
  
    if((INCOMING_AUTH_REQUIRED && authenticated) || !INCOMING_AUTH_REQUIRED){
      int jsonStart = readString.indexOf("{");
      String json = readString.substring(jsonStart);
      Serial.println(json);
      Serial.println("");
  
      //Process json
      processJson(json);
      
      //client.print(F("HTTP/1.1 200 OK\r\nContent-Type: application/json; charset=utf-8\r\n\r\n{\"method\": \"Ping\"}"));
      //The server really wants a response, I dont know what, so give it junk to avoid red errors :D
      client.print(F("HTTP/1.1 200 OK\r\nContent-Type: application/json; charset=utf-8\r\n\r\n{\"args\": [\"suffix\",\"Something\"],\"kwargs\": {\"prefix\":\"KWARGS\",\"payload\":\"None\",\"source\":\"arduino\"}, \"method\": \"TriggerEvent\"}"));
      
      // The client will actually be *flushed* then disconnected
      // when the function returns and 'client' object is destroyed (out-of-scope)
      // flush = ensure written data are received by the other side
      Serial.println(F("Disconnecting from client"));
      client.flush();
      client.stop();
    }
  }

  //Handle incoming RF messages
  if (mySwitch.available()) {
    String decimal = String(mySwitch.getReceivedValue());
    unsigned int length = mySwitch.getReceivedBitlength();
    String delay = String(mySwitch.getReceivedDelay());
    unsigned int* raw = mySwitch.getReceivedRawdata();
    String rawStr =  "";
    Serial.print("Raw data: ");
    for (unsigned int i=0; i<= length*2; i++) {
      rawStr = rawStr + String(raw[i]);
      rawStr = rawStr + ",";
    }
    String protocol = String(mySwitch.getReceivedProtocol());
    mySwitch.resetAvailable();
    
    //Send data to eventghost in json format
    post("TriggerEvent&prefix=RF433MHZ&suffix=Incoming&payload={\"decimal\": \""+decimal+"\", \"length\": \""+String(length)+"\", \"delay\": \""+delay+"\", \"raw\": \""+rawStr+"\", \"protocol\": \""+protocol+"\"}");
  }
}

//In EventGhost, use action:
//Webserver:  Send event to another EventGhost
//Event:      decimalCode:13964556,bitLength:24
void processJson(String jsonData)
{
  Serial.println("Process jsonData");
  //Not really json processing, just looking for a string, to send RF commands to my lights
  //void RCSwitch::send(unsigned long code, unsigned int length) {

  long decimalCode = extractValueFromKey(jsonData, "decimalCode").toInt();
  int bitLength = extractValueFromKey(jsonData, "bitLength").toInt();
  int pulseLength = extractValueFromKey(jsonData, "pulseLength").toInt();
  int repeat = extractValueFromKey(jsonData, "repeat").toInt();

  if (repeat == -1)
    repeat = 1;
    
  if(pulseLength > -1)
  {
    mySwitch.setPulseLength(pulseLength);
  }

  if(decimalCode != -1 && bitLength != -1){
    for (unsigned int i=0; i < repeat; i++) {
      mySwitch.send(decimalCode, bitLength);
    }
  }
}

String extractValueFromKey(String Data, String key)
{
  int startKeyValue = Data.indexOf(key) + key.length() + 1;
  if(Data.indexOf(key) != -1)
  {
    int endKeyValue = -1;
    endKeyValue = Data.indexOf(",", startKeyValue );
    if(endKeyValue == -1 || Data.substring(endKeyValue -1, endKeyValue).equals("]")){
      endKeyValue = Data.indexOf("\"", startKeyValue );
    }
  
    String keyValue = Data.substring(startKeyValue,endKeyValue);
    Serial.print(key);
    Serial.print(".Value=");
    Serial.println(keyValue);
    return keyValue;
  }

  return "-1";
}

//HTTP JSON POST to eventghost server
void post(String jsonData)
{
  HTTPClient http;
  //example "http://admin:admin@192.168.0.60:8888/"
  String url = "";
  if(OUTGOING_AUTH_REQUIRED){url = "http://" + outgoing_AUTH_USERNAME + ":" + outgoing_AUTH_PASSWORD + "@" + outgoing_IP + ":" + outgoing_PORT + "/";}
  else  {    url = "http://" + outgoing_IP + ":" + outgoing_PORT + "/";  }
  http.begin(url);
  http.addHeader("User-Agent", "Arduino/1.0");
  http.addHeader("Host", outgoing_IP);
  http.addHeader("Content-Type", "application/json; charset=utf-8"); //http.addHeader("Content-Type", "text/plain");
  int httpCode = http.POST(jsonData);
  String payload = http.getString();    //Get the response payload
  Serial.println(httpCode);   //Print HTTP return code
  Serial.println(payload);    //Print request response payload
  http.end();  //Close connection
}
